@@ -148,4 +148,17 @@ def custom_procrustes(X, Y, reflection='best'):
 # Here A and B have same shape n x d, T is d x d and v is 1 x d
 def procrustes(A, B):
     T, c = custom_procrustes(B,A)
-    return T, c
\ No newline at end of file
+    return T, c
+
+def ixmax(x, k=0, idx=None):
+    col = x[idx, k] if idx is not None else x[:, k]
+    z = np.where(col == col.max())[0]
+    return z if idx is None else idx[z]
+
+def lexargmax(x):
+    idx = None
+    for k in range(x.shape[1]):
+        idx = ixmax(x, k, idx)
+        if len(idx) < 2:
+            break
+    return idx[0]
\ No newline at end of file
